#!/usr/bin/env python

# http://en.wikipedia.org/wiki/Stop_codon
# http://en.wikipedia.org/wiki/Escherichia_coli
# http://en.wikipedia.org/wiki/Open_reading_frame
# http://nl.wikipedia.org/wiki/Genetische_code

import ghmm
import sys
import csv
import string


def ecoli_hmm(seq):
    """Try to find genes inside e sequence using a HMM"""
    # Model 4 bases A C G T and unknown state N
    sigma = ghmm.Alphabet(['a', 'c', 'g', 't', 'n' ])
    print sigma
   
    # XXX: Proper values, based of statistics
    # The transition matrix A is chosen such that it reflects the statistics
    # Part one will try to get us into a gene, part two will tell us when to
    # get out of it
    A = [[0.6,0.4], [0.3, 0.7]]

    # The emission probabilities matrix is modeled after the statistics with a
    # total of 1
    etogene = [ 0.1, 0.1, 0.1, 0.1, 0.6 ]
    eingene = [ 0.2, 0.3, 0.3, 0.2, 0.1 ]

    B = [etogene,eingene]
    
    # Initial distribution favors outside
    pi  = [0.9, 0.1]

    m = ghmm.HMMFromMatrices(sigma,ghmm.DiscreteDistribution(sigma),A ,B, pi)
    print "Initial HMM", m

    obs_seq = m.sampleSingle(20)
    print "Observation sequence : ", obs_seq
    obs = map(sigma.external, obs_seq)
    print "Observations         : ", ''.join(obs)

    # Start codons
    # atg, (small chance) gtg
    # Stop codons
    # taa, tga
    # tag
    # lend = Left End
    # rend = Right End

    # Training
    # A -- T, G -- C
    start_condons = ['atg', 'gtg']
    stop_condons = ['taa', 'tga', 'tag']
    dna_flip = string.maketrans('atgc','tacg')
    edl_file = 'data/edl_genes.csv'
    reader = csv.reader(open(edl_file,"rU"))
    reader.next()
    try:
        for row in reader:
            (featureType, zNumber, contig, lend, rend, orientation, 
            segmentType, oIslandNumber, geneName, note, function, product, 
            translationNotes) = row
            lend = int(lend) - 1
            rend = int(rend)
            # Make a forward orientation as positive
            if orientation == '>':
                order = True
            else:
                order = False

            if order:
                start_codon = seq[lend:lend+3]
                stop_codon = seq[rend-3:rend]
            else:
                start_codon = seq[rend-3:rend].translate(dna_flip)[::-1]
                stop_codon = seq[lend:lend+3].translate(dna_flip)[::-1]

            print "%6s, %s, %s, %s, %s, %s," % (geneName,lend,rend, start_codon, stop_codon, orientation),

            print "%s, %s" % (start_codon in start_condons, stop_codon in stop_condons)
            if not start_codon in start_condons or not stop_codon in stop_condons:
                return
    except csv.Error, e:
        sys.exit('file %s, line %d: %s' % (filename, reader.line_num, e))


        
    test_seq=ghmm.EmissionSequence(sigma,list(seq[0:(len(seq) / 3)]))
    v = m.viterbi(test_seq)
    #print "fairness of test_seq: ", v

    
    # Validation
    val_seq=ghmm.EmissionSequence(sigma,list(seq[10:2000]))
    v = m.viterbi(val_seq)
    #print "Test sequence: ", v

    # XXX: Results



if __name__ == "__main__":
    # Load data
    try:
        handle = open(sys.argv[1],"rU")
    except IndexError:
        print "Usage %s <data_file>" % (sys.argv[0])
        sys.exit(64)
    except IOError:
        print "Unable to open '%s'" % (sys.argv[1])
        sys.exit(64)
    
    seq = handle.read()
    handle.close()
    ecoli_hmm(seq)
